package dk.aau.cs.io;

import dk.aau.cs.gui.NameGenerator;
import dk.aau.cs.model.tapn.*;
import dk.aau.cs.model.tapn.Colored.*;
import dk.aau.cs.model.tapn.Colored.Color;
import dk.aau.cs.model.tapn.Colored.Expressions.*;
import dk.aau.cs.util.FormatException;
import dk.aau.cs.util.Require;
import dk.aau.cs.util.Tuple;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;
import pipe.dataLayer.DataLayer;
import pipe.dataLayer.NetType;
import pipe.dataLayer.TAPNQuery;
import pipe.dataLayer.Template;
import pipe.gui.DrawingSurfaceImpl;
import pipe.gui.Pipe;
import pipe.gui.Zoomer;
import pipe.gui.graphicElements.*;
import pipe.gui.graphicElements.cpn.ColoredInputArcComponent;
import pipe.gui.graphicElements.cpn.ColoredOutputArcComponent;
import pipe.gui.graphicElements.cpn.ColoredPlaceComponent;
import pipe.gui.graphicElements.cpn.ColoredTransitionComponent;
import pipe.gui.graphicElements.tapn.*;
import pipe.gui.handler.*;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.awt.*;
import java.io.*;
import java.util.*;

public class PNMLoader {
	
	enum GraphicsType { Position, Offset }

	private DrawingSurfaceImpl drawingSurface;
	private NameGenerator nameGenerator = new NameGenerator();
	private IdResolver idResolver = new IdResolver();
	private HashSet<String> arcs = new HashSet<String>();
	private HashMap<String, TimedPlace> places = new HashMap<String, TimedPlace>();
	private HashMap<String, TimedTransition> transitions = new HashMap<String, TimedTransition>();
	private HashMap<String, ColorType> colortypes = new HashMap<String, ColorType>();
	private HashMap<String, Variable> variables = new HashMap<String, Variable>();
	
	//If the net is too big, do not make the graphics
	private int netSize = 0;
	private int maxNetSize = 4000;

	public PNMLoader(DrawingSurfaceImpl drawingSurface) {
		this.drawingSurface = drawingSurface;
	}
	
	public LoadedModel load(File file) throws FormatException{
		try{
			return load(new FileInputStream(file));
		} catch (FileNotFoundException e){
			return null;
		} catch (NullPointerException e){
			throw new FormatException("the PNML file cannot be parsed due to syntax errors");
		}
	}
	
	public LoadedModel load(InputStream file) throws FormatException{
		Document doc = loadDocument(file);

		return parse(doc);
	}

	private Document loadDocument(InputStream file) {
		try {
			DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			return builder.parse(file);
		} catch (ParserConfigurationException e) {
			return null;
		} catch (SAXException e) {
			return null;
		} catch (IOException e) {
			return null;
		}
	}
	
	private LoadedModel parse(Document doc) throws FormatException {
		idResolver.clear();
		
		//We assume there is only one net per file (this is what we call a TAPN Network) 
		Node pnmlElement = doc.getElementsByTagName("pnml").item(0);
		Node netNode = getFirstDirectChild(pnmlElement, "net");

		String name = getTAPNName(netNode);
		boolean isColored = (getFirstDirectChild(netNode, "declaration") != null);

		TimedArcPetriNetNetwork network;
        TimedArcPetriNet tapn;

		if (isColored) {
			network = new ColoredArcPetriNetNetwork();
            tapn = new ColoredArcPetriNet(name);
		} else {
			network = new TimedArcPetriNetNetwork();
            tapn = new TimedArcPetriNet(name);
		}

		tapn.setCheckNames(false);
		network.add(tapn);
		nameGenerator.add(tapn);

		//We assume there is only one page pr. file (this is what we call a net)
        Template template;
		if (isColored) {
            template = new Template(tapn, new DataLayer(NetType.CPN), new Zoomer());
        }
        else {
            template = new Template(tapn, new DataLayer(NetType.TAPN), new Zoomer());
        }

	
		parseTimedArcPetriNet(netNode, tapn, template);
		
		network.setPaintNet(isNetDrawable());
		tapn.setCheckNames(true);
		return new LoadedModel(network, Arrays.asList(template), new ArrayList<TAPNQuery>());
	}

	private String getTAPNName(Node netNode) {
		if(netNode == null || !(netNode instanceof Element)){
			return nameGenerator.getNewTemplateName();
		}
		String result = null;

		Node name =  getFirstDirectChild(netNode, "name");
		if(name != null){
			result = getFirstDirectChild(name, "text").getTextContent();
		}
		
		if(name == null || name.equals("")){
			return nameGenerator.getNewTemplateName();
		}
		
		return NamePurifier.purify(result);
	}

	private void parseTimedArcPetriNet(Node netNode, TimedArcPetriNet tapn, Template template) throws FormatException {
		//Parse declarations first in case of a colored net

		if (tapn.parentNetwork().isColored()) {
			Node declarations = getFirstDirectChild(netNode, "declaration");
			parseDeclarations(declarations, tapn, template);
		}

		//We assume there is only one page pr. file (this is what we call a net) 
		Node node = getFirstDirectChild(netNode, "page").getFirstChild();
		Node first = node;
		
		//Calculate netsize
		while(node != null){
			netSize += 1;
			node = skipWS(node.getNextSibling());
		}
		

		node = first;
		//We parse the places and transitions first
		while(node != null){
			String tag = node.getNodeName();
			if(tag.equals("place")){
				parsePlace(node, tapn, template);
			} else if(tag.equals("transition")){
				parseTransition(node, tapn, template);
			}
			node = skipWS(node.getNextSibling());
		}
		
		//We parse the transitions last, as we need the places and transitions it refers to
		node = first;
		while(node != null){
			String tag = node.getNodeName();
			if(tag.equals("arc")){
				parseArc(node, tapn, template);
			}
			skipWS(node = node.getNextSibling());
		}
	}

	private void parseDeclarations(Node node, TimedArcPetriNet tapn, Template template) throws FormatException {
		if(node == null || !(node instanceof Element)){
			return;
		}

		Node child = skipWS(node.getFirstChild());
		while(child != null){
			String childName = child.getNodeName();
			if (childName.equals("namedsort")){
				parseNamedSort(child, tapn, template);
			} else if (childName.equals("variabledecl")){
				String id = getAttribute(child, "id").getNodeValue();
				String name = getAttribute(child, "name").getNodeValue();
				ColorType ct = parseUserSort(child);
				Variable var = new Variable(name, id, ct);
				Require.that(variables.put(id, var) == null, "the id " + id + ", was already used");
				tapn.parentNetwork().add(var);
			} else {
				parseDeclarations(child, tapn, template);
			}

			child = skipWS(child.getNextSibling());
		}
	}

	private void parseNamedSort(Node node, TimedArcPetriNet tapn, Template template) throws FormatException {
		Node type = skipWS(node.getFirstChild());
		String typetag = type.getNodeName();

		String id = getAttribute(node, "id").getNodeValue();
		String name = getAttribute(node, "name").getNodeValue();
		if (typetag.equals("productsort")) {
			ProductType pt = new ProductType(name, id);
			Node typechild = skipWS(type.getFirstChild());
			while (typechild != null) {
				if (typechild.getNodeName().equals("usersort")) {
					String constituent = getAttribute(typechild, "declaration").getNodeValue();
					pt.addType(colortypes.get(constituent));
				}
				typechild = skipWS(typechild.getNextSibling());
			}
			Require.that(colortypes.put(id, pt) == null, "the name " + name + ", was already used");
			tapn.parentNetwork().add(pt);

		} else {
			ColorType ct = new ColorType(id, name);
			if (typetag.equals("dot")) {
				ct.addColor("dot");
			} else {
				Node typechild = skipWS(type.getFirstChild());
				while (typechild != null) {
					Node dotId = getAttribute(typechild, "id");
					if (dotId != null) {
						ct.addColor(dotId.getNodeValue());
						typechild = skipWS(typechild.getNextSibling());
					} else {
						throw new FormatException(String.format("No id found on %s\n", typechild.getNodeName()));
					}
				}
			}
			Require.that(colortypes.put(id, ct) == null, "the name " + name + ", was already used");
			tapn.parentNetwork().add(ct);
		}
	}

	private ColorType parseUserSort(Node node) throws FormatException {
		if (node != null && node instanceof Element) {
			Node child = skipWS(node.getFirstChild());
			while (child != null) {
				String name = child.getNodeName();
				if (name.equals("usersort")) {
					Node decl = getAttribute(child, "declaration");
					return colortypes.get(decl.getNodeValue());
				} else if (name.matches("structure|type|subterm")) {
					return parseUserSort(child);
				}
				child = skipWS(child.getNextSibling());
			}
		}
		throw new FormatException(String.format("Could not parse %s as an usersort\n", node.getNodeName()));
	}

	private void parsePlace(Node node, TimedArcPetriNet tapn, Template template) throws FormatException {
		if(node == null || !(node instanceof Element)){
			return;
		}

		Name name = parseName(getFirstDirectChild(node, "name"));
		if(name == null){
			name = new Name(nameGenerator.getNewPlaceName(template.model()));
		}
		Point position = parseGraphics(getFirstDirectChild(node, "graphics"), GraphicsType.Position);
		String id = NamePurifier.purify(((Element) node).getAttribute("id"));
		InitialMarking marking = null;
		ArcExpression colorMarking = null;
		Point markingOffset = null;
		TimedPlace place;
		if (tapn.parentNetwork().isColored()) {

			Node typeNode = getFirstDirectChild(node, "type");
			ColorType colorType = null;
			if(typeNode != null) {
				colorType = parseUserSort(typeNode);
			}

				Node markingNode = getFirstDirectChild(node, "hlinitialMarking");
				if (markingNode == null) {
					markingOffset = new Point();
					colorMarking = null;
				} else {
					markingOffset = parseGraphics(getFirstDirectChild(markingNode, "graphics"), GraphicsType.Offset);
					Node markingExpression = getFirstDirectChild(markingNode, "structure");
					colorMarking = parseArcExpression(markingExpression);
				}
					place = new LocalColoredPlace(id, colorType);

		} else {
			marking = parseMarking(getFirstDirectChild(node, "initialMarking"));

			place = new LocalTimedPlace(id, new TimeInvariant(false, new Bound.InfBound()));
		}
		Require.that(places.put(id, place) == null && !transitions.containsKey(id), 
				"The name: " + id + ", was already used");
		tapn.add(place);
		
		if(isNetDrawable()){
			//We parse the id as both the name and id as in tapaal name = id, and name/id has to be unique
			TimedPlaceComponent placeComponent;
			if (tapn.parentNetwork().isColored()) {
				placeComponent = new ColoredPlaceComponent(position.getX(), position.getY(), id, id, name.point.getX(),
						name.point.getY(), markingOffset.x, markingOffset.y);
			} else {
				placeComponent = new TimedPlaceComponent(position.getX(), position.getY(), id, id,
						name.point.getX(), name.point.getY(), marking.marking, marking.point.x, marking.point.y, 0);
			}
			placeComponent.setUnderlyingPlace(place);
			template.guiModel().addPetriNetObject(placeComponent);
			addListeners(placeComponent, template);
		}
		
		idResolver.add(tapn.name(), id, id);

		if (tapn.parentNetwork().isColored()) {
			if (colorMarking != null) {
				ExpressionContext context = new ExpressionContext(new HashMap<String, Color>(), colortypes);
				ColorMultiset cm = colorMarking.eval(context);
				for (ColoredToken ct : cm.getTokens(place)) {
					tapn.parentNetwork().marking().add(ct);
				}
			}
		} else {
			for (int i = 0; i < marking.marking; i++) {
				tapn.parentNetwork().marking().add(new TimedToken(place));
			}
		}
	}
	
	private InitialMarking parseMarking(Node node) {
		if(node == null || !(node instanceof Element)){
			return new InitialMarking();
		}
		
		Point offset = parseGraphics(getFirstDirectChild(node, "graphics"), GraphicsType.Offset);
		
		int marking = Integer.parseInt(getFirstDirectChild(node, "text").getTextContent());
		
		return new InitialMarking(marking, offset);
	}

	private void parseTransition(Node node, TimedArcPetriNet tapn, Template template) throws FormatException {
		if(node == null || !(node instanceof Element)){
			return;
		}
		
		Point position = parseGraphics(getFirstDirectChild(node, "graphics"), GraphicsType.Position);
		Name name = parseName(getFirstDirectChild(node, "name"));
		if(name == null){
			name = new Name(nameGenerator.getNewTransitionName(template.model()));
		}
		String id = NamePurifier.purify(((Element) node).getAttribute("id"));

		TimedTransition transition;
		if (tapn.parentNetwork().isColored()) {
			GuardExpression guardexp = null;
			Node guardnode = getFirstDirectChild(node, "condition");
			if (guardnode != null) {
				guardexp = parseGuardExpression(getFirstDirectChild(guardnode, "structure"));
			}
			transition = new ColoredTransition(id, guardexp);
		} else {
			transition = new TimedTransition(id);
		}

		Require.that(transitions.put(id, transition) == null && !places.containsKey(id), 
				"The id: " + id + ", was already used");
		tapn.add(transition);
		
		if(isNetDrawable()){
			//We parse the id as both the name and id as in tapaal name = id, and name/id has to be unique
			TimedTransitionComponent transitionComponent;
			if (tapn.parentNetwork().isColored()) {
				transitionComponent = new ColoredTransitionComponent(position.getX(), position.getY(), id, id,
						name.point.getX(), name.point.getY(), true, false, 0, 0);
			} else {
				transitionComponent = new TimedTransitionComponent(position.getX(), position.getY(), id, id,
						name.point.getX(), name.point.getY(), true, false, 0, 0);
			}
			transitionComponent.setUnderlyingTransition(transition);
			template.guiModel().addPetriNetObject(transitionComponent);
			addListeners(transitionComponent, template);
			template.guiModel().repaintAll(false);
		}
		idResolver.add(tapn.name(), id, id);
	}
	
	private void parseArc(Node node, TimedArcPetriNet tapn, Template template) throws FormatException {
		if(node == null || !(node instanceof Element)){
			return;
		}
		
		Element element = (Element) node;
		
		String id = element.getAttribute("id");
		String sourceId = NamePurifier.purify(element.getAttribute("source"));
		String targetId = NamePurifier.purify(element.getAttribute("target"));
		String type = element.getAttribute("type");
		
		String sourceName = idResolver.get(template.model().name(), sourceId);
		String targetName = idResolver.get(template.model().name(), targetId);
		
		TimedPlace sourcePlace = places.get(sourceName);
		TimedPlace targetPlace = places.get(targetName);
		
		TimedTransition sourceTransition = transitions.get(sourceName);
		TimedTransition targetTransition = transitions.get(targetName);
		
		PlaceTransitionObject source = template.guiModel().getPlaceTransitionObject(sourceName);
		PlaceTransitionObject target = template.guiModel().getPlaceTransitionObject(targetName);
		
		//Inscription
		int weight = 1;
		ArcExpression expr = null;
		if (tapn.parentNetwork().isColored()) {
			Node hlinscription = getFirstDirectChild(node, "hlinscription");
			if(hlinscription != null) {
				hlinscription = getFirstDirectChild(hlinscription, "structure");
			}
			if (hlinscription != null) {
				expr = parseArcExpression(hlinscription);
			}
		} else {
			Node inscription = getFirstDirectChild(node, "inscription");
			if(inscription != null){
				Node text = getFirstDirectChild(inscription, "text");
				if(text != null){
					String weightString = text.getTextContent().trim();
					try{
						if(weightString != null)
							weight = Integer.parseInt(weightString);
					} catch(NumberFormatException e) {}
				}
			}
		}
		
		int _startx = 0, _starty = 0, _endx = 0, _endy = 0;
		
		if(isNetDrawable()){
			// add the insets and offset
			_startx = source.getX() + source.centreOffsetLeft();
			_starty = source.getY() + source.centreOffsetTop();

			_endx = target.getX() + target.centreOffsetLeft();
			_endy = target.getY() + target.centreOffsetTop();
		}
		
		Arc tempArc;
		
		if(type != null && type.equals("inhibitor")) {
			tempArc = parseAndAddTimedInhibitorArc(id, sourcePlace, targetTransition, source, target, weight, _startx, _starty, _endx, _endy, template);
		} else if(sourcePlace != null && targetTransition != null) {
			tempArc = parseInputArc(id, sourcePlace, targetTransition, source, target, weight, expr, _startx, _starty, _endx, _endy, tapn, template);
		} else if(sourceTransition != null && targetPlace != null) {
			tempArc = parseOutputArc(id,  sourceTransition, targetPlace, source, target, weight, expr, _startx, _starty, _endx, _endy, tapn, template);
		} else {
			throw new FormatException("Arcs must be only between places and transitions");
		}
		
		if(isNetDrawable()) parseArcPath(element, tempArc);
	}

	private ArcExpression parseArcExpression(Node node) throws FormatException {
		String name = node.getNodeName();
		if (name.equals("numberof")) {
			return parseNumberOfExpression(node);
		} else if (name.equals("add")) {
			Vector<ArcExpression> constituents = new Vector<ArcExpression>();

			Node child = skipWS(node.getFirstChild());
			while (child != null) {
				ArcExpression subterm = parseArcExpression(child);
				constituents.add(subterm);
				child = skipWS(child.getNextSibling());
			}
			return new AddExpression(constituents);
		} else if (name.equals("subtract")) {
			Node headchild = skipWS(node.getFirstChild());
			ArcExpression headexp = parseArcExpression(headchild);

			Node nextchild = skipWS(headchild.getNextSibling());
			while (nextchild != null) {
				ArcExpression nextexp = parseArcExpression(nextchild);
				headexp = new SubtractExpression(headexp, nextexp);
				nextchild = skipWS(nextchild.getNextSibling());
			}
			return headexp;
		} else if (name.equals("scalarproduct")) {
			Node scalar = skipWS(node.getFirstChild());
			Integer scalarval = parseNumberConstantExpression(scalar);

			Node child = skipWS(scalar.getNextSibling());
			ArcExpression childexp = parseArcExpression(child);

			return new ScalarProductExpression(scalarval, childexp);

		} else if (name.equals("all")) {
			Node parent = node.getParentNode();
			return parseNumberOfExpression(parent);
		} else if (name.matches("subterm|structure")) {
			Node child = skipWS(node.getFirstChild());
			return parseArcExpression(child);
		} else {
			throw new FormatException(String.format("Could not parse %s as an arc expression\n", name));
		}
	}

	private NumberOfExpression parseNumberOfExpression(Node node) throws FormatException {
		Node number = skipWS(node.getFirstChild());
		//The number constant may be omitted.
		//In that case, this parsing returns null.
		Integer numberval = parseNumberConstantExpression(number);
		Node subnode;
		if (numberval != null) {
			//The subexpression comes after the number constant.
			subnode = skipWS(number.getNextSibling());
		} else {
			//The number we read was actually the subexpression.
			subnode = number;
			numberval = 1;
		}
		//Try to parse subexpression as all expression
		AllExpression subexp = parseAllExpression(subnode);

		if (subexp != null) {
			return new NumberOfExpression(numberval, subexp);
		} else {
			Vector<ColorExpression> colorexps = new Vector<ColorExpression>();
			while (subnode != null) {
				ColorExpression colorexp = parseColorExpression(subnode);
				colorexps.add(colorexp);
				subnode = skipWS(subnode.getNextSibling());
			}
			return new NumberOfExpression(numberval, colorexps);
		}
	}

	private Integer parseNumberConstantExpression(Node node) {
		String name = node.getNodeName();
		if (name.equals("numberconstant")) {
			String value = getAttribute(node, "value").getNodeValue();
			return Integer.valueOf(value);
		} else if (name.equals("subterm")) {
			Node child = skipWS(node.getFirstChild());
			return parseNumberConstantExpression(child);
		} else {
			return null;
		}
	}

	private AllExpression parseAllExpression(Node node) throws FormatException {
		String name = node.getNodeName();
		if (name.equals("all")) {
			ColorType ct = parseUserSort(node);
			return new AllExpression(ct);
		} else if (name.equals("subterm")) {
			Node child = skipWS(node.getFirstChild());
			return parseAllExpression(child);
		} else {
			return null;
		}
	}

	private ColorExpression parseColorExpression(Node node) throws FormatException {
		String name = node.getNodeName();
		if (name.equals("dotconstant")) {
			return new DotConstantExpression();
		} else if (name.equals("variable")) {
			String varname = getAttribute(node, "refvariable").getNodeValue();
			Variable var = variables.get(varname);
			return new VariableExpression(var);
		} else if (name.equals("useroperator")) {
			String colorname = getAttribute(node, "declaration").getNodeValue();
			Color color = getColor(colorname);
			return new UserOperatorExpression(color);
		} else if (name.equals("successor")) {
			Node child = skipWS(node.getFirstChild());
			ColorExpression childexp = parseColorExpression(child);
			return new SuccessorExpression(childexp);
		} else if (name.equals("predecessor")) {
			Node child = skipWS(node.getFirstChild());
			ColorExpression childexp = parseColorExpression(child);
			return new PredecessorExpression(childexp);
		} else if (name.equals("tuple")) {
			Vector<ColorExpression> colorexps = new Vector<ColorExpression>();

			Node child = skipWS(node.getFirstChild());
			while (child != null) {
				ColorExpression colorexp = parseColorExpression(child);
				colorexps.add(colorexp);
				child = skipWS(child.getNextSibling());
			}
			return new TupleExpression(colorexps);
		} else if (name.matches("subterm|structure")) {
			Node child = skipWS(node.getFirstChild());
			return parseColorExpression(child);
		} else {
			throw new FormatException(String.format("Could not parse %s as an color expression\n", name));
		}
	}

	private GuardExpression parseGuardExpression(Node node) throws FormatException {
		String name = node.getNodeName();
		if (name.matches("lt|lessthan")) {
			Tuple<ColorExpression, ColorExpression> subexps = parseLRColorExpressions(node);
			return new LessThanExpression(subexps.value1(), subexps.value2());
		} else if (name.matches("gt|greaterthan")) {
			Tuple<ColorExpression, ColorExpression> subexps = parseLRColorExpressions(node);
			return new GreaterThanExpression(subexps.value1(), subexps.value2());
		} else if (name.matches("leq|lessthanorequal")) {
			Tuple<ColorExpression, ColorExpression> subexps = parseLRColorExpressions(node);
			return new LessThanEqExpression(subexps.value1(), subexps.value2());
		} else if (name.matches("geq|greaterthanorequal")) {
			Tuple<ColorExpression, ColorExpression> subexps = parseLRColorExpressions(node);
			return new GreaterThanEqExpression(subexps.value1(), subexps.value2());
		} else if (name.matches("eq|equality")) {
			Tuple<ColorExpression, ColorExpression> subexps = parseLRColorExpressions(node);
			return new EqualityExpression(subexps.value1(), subexps.value2());
		} else if (name.matches("neq|inequality")) {
			Tuple<ColorExpression, ColorExpression> subexps = parseLRColorExpressions(node);
			return new InequalityExpression(subexps.value1(), subexps.value2());
		} else if (name.equals("not")) {
			Node child = skipWS(node.getFirstChild());
			GuardExpression childexp = parseGuardExpression(child);
			return new NotExpression(childexp);
		} else if (name.equals("and")) {
			Tuple<GuardExpression, GuardExpression> subexps = parseLRGuardExpressions(node);
			return new AndExpression(subexps.value1(), subexps.value2());
		} else if (name.equals("or")) {
			Tuple<GuardExpression, GuardExpression> subexps = parseLRGuardExpressions(node);
			return new OrExpression(subexps.value1(), subexps.value2());
		} else if (name.matches("subterm|structure")) {
			Node child = skipWS(node.getFirstChild());
			return parseGuardExpression(child);
		} else {
			throw new FormatException(String.format("Could not parse %s as a guard expression\n", name));
		}
	}

	private Tuple<ColorExpression,ColorExpression> parseLRColorExpressions(Node node) throws FormatException {
		Node left = skipWS(node.getFirstChild());
		ColorExpression leftexp = parseColorExpression(left);
		Node right = skipWS(left.getNextSibling());
		ColorExpression rightexp = parseColorExpression(right);
		return new Tuple<ColorExpression, ColorExpression>(leftexp, rightexp);
	}

	private Tuple<GuardExpression,GuardExpression> parseLRGuardExpressions(Node node) throws FormatException {
		Node left = skipWS(node.getFirstChild());
		GuardExpression leftexp = parseGuardExpression(left);
		Node right = skipWS(left.getNextSibling());
		GuardExpression rightexp = parseGuardExpression(right);
		return new Tuple<GuardExpression, GuardExpression>(leftexp, rightexp);
	}
	
	private void parseArcPath(Element arc, Arc tempArc) {
		Element element = (Element) getFirstDirectChild(arc, "graphics");
		if(element == null) return;
		NodeList nodelist = element.getElementsByTagName("position");
		if (nodelist.getLength() > 0) {
			for (int i = 0; i < nodelist.getLength(); i++) {
				Node node = nodelist.item(i);
				if (node instanceof Element) {
					Element position = (Element) node;
					if ("position".equals(position.getNodeName())) {
						String arcTempX = position.getAttribute("x");
						String arcTempY = position.getAttribute("y");

						float arcPointX = Float.valueOf(arcTempX).floatValue();
						float arcPointY = Float.valueOf(arcTempY).floatValue();
						arcPointX += Pipe.ARC_CONTROL_POINT_CONSTANT + 1;
						arcPointY += Pipe.ARC_CONTROL_POINT_CONSTANT + 1;
						
						//We add the point at i+1 as the starting and end points of 
						//the arc is already in the path as point number 0 and 1
						tempArc.getArcPath().addPoint(i+1,arcPointX, arcPointY, false);
					}
				}
			}
		}
	}

	private Name parseName(Node node){
		if(node == null || !(node instanceof Element)){
			return null;
		}
		Point offset = parseGraphics(getFirstDirectChild(node, "graphics"), GraphicsType.Offset);
		
		String name = getFirstDirectChild(node, "text").getTextContent();
		if(name == null || name.equals("")){
			return null;
		}
		
		name = NamePurifier.purify(name);
		return new Name(name, offset);
	}
	
	private Point parseGraphics(Node node, GraphicsType type){
		if(node == null || !(node instanceof Element)){
			if(type == GraphicsType.Offset)
				return new Point(0, -10);
			else 
				return new Point(100, 100);
		}
		
		Element offset = (Element)getFirstDirectChild(node, type == GraphicsType.Offset ? "offset" : "position");
		
		String x = offset.getAttribute("x");
		String y = offset.getAttribute("y");
		
                int xd = Math.round(Float.valueOf(x).floatValue());
                int yd = Math.round(Float.valueOf(y).floatValue());
               
		return new Point(xd, yd);
	}

	private class Name{
		String name;
		Point point;
		
		public Name(String newPlaceName) {
			this(newPlaceName, new Point());
		}
		
		public Name(String name, Point p) {
			this.name = name;
			this.point = p;
		}

		@Override
		public String toString() {
			return name + ";" + point;
		}
	}
	
	private class InitialMarking{
		int marking;
		Point point;
		
		public InitialMarking() {
			this(0, new Point());
		}
			
		
		public InitialMarking(int marking, Point p) {
			this.marking = marking;
			this.point = p;
		}
		
		@Override
		public String toString() {
			return marking + ";" + point;
		}
	}
	
	private TimedInputArcComponent parseInputArc(String arcId, TimedPlace place, TimedTransition transition, PlaceTransitionObject source,
			PlaceTransitionObject target, int weight, ArcExpression expression, int _startx, int _starty, int _endx,
			int _endy, TimedArcPetriNet tapn, Template template) throws FormatException {

		TimedInputArc inputArc;
		if (tapn.parentNetwork().isColored()) {
			inputArc = new ColoredInputArc(place, transition,
					new ArrayList<ColoredTimeInterval>(){{add(ColoredTimeInterval.ZERO_INF_DYN_COLOR(Color.STAR_COLOR));}}, expression);
		} else {
			inputArc = new TimedInputArc(place, transition, TimeInterval.ZERO_INF, new IntWeight(weight));
		}
		
		Require.that(places.containsKey(inputArc.source().name()),	"The source place must be part of the petri net.");
		Require.that(transitions.containsKey(inputArc.destination().name()), "The destination transition must be part of the petri net");
		if(!arcs.add(inputArc.source().name() + "-in-" + inputArc.destination().name())) {
			throw new FormatException("Multiple arcs between a place and a transition is not allowed");
		}
		
		TimedInputArcComponent arc = null;
		
		if(isNetDrawable()){
			if (tapn.parentNetwork().isColored()) {
				arc = new ColoredInputArcComponent(new ColoredOutputArcComponent(_startx, _starty, _endx, _endy, source, target, weight, expression, arcId,
						false));
			} else {
				arc = new TimedInputArcComponent(new TimedOutputArcComponent(
						_startx, _starty, _endx, _endy, source, target, weight, arcId,
						false));
			}

			arc.setUnderlyingArc(inputArc);

			template.guiModel().addPetriNetObject(arc);
			addListeners(arc, template);

			source.addConnectFrom(arc);
			target.addConnectTo(arc);
		}
		
		template.model().add(inputArc);
		
		return arc;

		
	}
	
	private Arc parseOutputArc(String arcId, TimedTransition transition, TimedPlace place, PlaceTransitionObject source,
			PlaceTransitionObject target, int weight, ArcExpression expression, int _startx, int _starty, int _endx,
			int _endy, TimedArcPetriNet tapn, Template template) throws FormatException {

		TimedOutputArc outputArc;
		if (tapn.parentNetwork().isColored()) {
			outputArc = new ColoredOutputArc(transition, place, expression);
		} else {
			outputArc = new TimedOutputArc(transition, place, new IntWeight(weight));
		}
		
		Require.that(places.containsKey(outputArc.destination().name()), "The destination place must be part of the petri net.");
		Require.that(transitions.containsKey(outputArc.source().name()), "The source transition must be part of the petri net");
		if(!arcs.add(outputArc.source().name() + "-out-" + outputArc.destination().name())) {
			throw new FormatException("Multiple arcs between a place and a transition is not allowed");
		}
		
		TimedOutputArcComponent arc = null;
		
		if(isNetDrawable()){
			if (tapn.parentNetwork().isColored()) {
				arc = new ColoredOutputArcComponent(_startx, _starty, _endx, _endy,
						source, target, weight,	expression, arcId, false);
			} else {
				arc = new TimedOutputArcComponent(_startx, _starty, _endx, _endy,
						source, target, weight,	arcId, false);
			}
			arc.setUnderlyingArc(outputArc);

			template.guiModel().addPetriNetObject(arc);
			addListeners(arc, template);

			source.addConnectFrom(arc);
			target.addConnectTo(arc);
		}
		
		template.model().add(outputArc);
		return arc;
	}
	
	private Arc parseAndAddTimedInhibitorArc(String arcId, TimedPlace place, TimedTransition transition, PlaceTransitionObject source,
			PlaceTransitionObject target, int weight, int _startx, int _starty, int _endx,
			int _endy, Template template) {
		TimedInhibitorArcComponent tempArc = new TimedInhibitorArcComponent(
				new TimedInputArcComponent(
						new TimedOutputArcComponent(_startx, _starty, _endx, _endy,	source, target, weight, arcId, false)
				),
				(""));
		
		TimedInhibitorArc inhibArc = new TimedInhibitorArc(place, transition, TimeInterval.ZERO_INF, new IntWeight(weight));

		tempArc.setUnderlyingArc(inhibArc);
		template.guiModel().addPetriNetObject(tempArc);
		addListeners(tempArc, template);
		template.model().add(inhibArc);

		source.addConnectFrom(tempArc);
		target.addConnectTo(tempArc);
		return tempArc;
	}
	
	private boolean isNetDrawable(){
		return netSize <= maxNetSize;
	}
	
	public static Node getFirstDirectChild(Node parent, String tagName){
		NodeList children = parent.getChildNodes();
		for(int i = 0; i < children.getLength(); i++){
			if(children.item(i).getNodeName().equals(tagName)){
				return children.item(i);
			}
		}
		return null;
	}

	private static Node getAttribute(Node node, String attribute) {
		return node.getAttributes().getNamedItem(attribute);
	}

	//skipWS
	//  Takes a given node and returns it if it is an element,
	//  otherwise returns its first element sibling.
	//  This allows skipping past whitespace nodes.
	private static Node skipWS(Node node) {
		if (node != null && !(node instanceof Element)) {
			return skipWS(node.getNextSibling());
		} else {
			return node;
		}
	}

	private Color getColor(String colorname) throws FormatException {
		for (ColorType ct : colortypes.values()) {
			for (Color c : ct) {
				if (c.getName().equals(colorname)) {
					return c;
				}
			}
		}
		throw new FormatException(String.format("The color \"%s\" was not declared\n", colorname));
	}
	
	private void addListeners(PetriNetObject newObject, Template template) {
		if (newObject != null) {
			if (newObject.getMouseListeners().length == 0) {
				if (newObject instanceof Place) {
					// XXX - kyrke
					if (newObject instanceof TimedPlaceComponent) {

						LabelHandler labelHandler = new LabelHandler(((Place) newObject).getNameLabel(), (Place) newObject);
						((Place) newObject).getNameLabel().addMouseListener(labelHandler);
						((Place) newObject).getNameLabel().addMouseMotionListener(labelHandler);
						((Place) newObject).getNameLabel().addMouseWheelListener(labelHandler);

						PlaceHandler placeHandler = new PlaceHandler(drawingSurface, (Place) newObject, template.guiModel(), template.model());
						newObject.addMouseListener(placeHandler);
						newObject.addMouseWheelListener(placeHandler);
						newObject.addMouseMotionListener(placeHandler);
					} else {

						LabelHandler labelHandler = new LabelHandler(((Place) newObject).getNameLabel(), (Place) newObject);
						((Place) newObject).getNameLabel().addMouseListener(labelHandler);
						((Place) newObject).getNameLabel().addMouseMotionListener(labelHandler);
						//((Place) newObject).getNameLabel().addMouseWheelListener(labelHandler);

						PlaceHandler placeHandler = new PlaceHandler(drawingSurface, (Place) newObject);
						newObject.addMouseListener(placeHandler);
						//newObject.addMouseWheelListener(placeHandler);
						newObject.addMouseMotionListener(placeHandler);

					}
				} else if (newObject instanceof Transition) {
					TransitionHandler transitionHandler;
					if (newObject instanceof TimedTransitionComponent) {
						transitionHandler = new TAPNTransitionHandler(drawingSurface, (Transition) newObject, template.guiModel(), template.model());
					} else {
						transitionHandler = new TransitionHandler(drawingSurface, (Transition) newObject);
					}

					LabelHandler labelHandler = new LabelHandler(((Transition) newObject).getNameLabel(), (Transition) newObject);
					((Transition) newObject).getNameLabel().addMouseListener(labelHandler);
					((Transition) newObject).getNameLabel().addMouseMotionListener(labelHandler);
					((Transition) newObject).getNameLabel().addMouseWheelListener(labelHandler);

					newObject.addMouseListener(transitionHandler);
					newObject.addMouseMotionListener(transitionHandler);
					newObject.addMouseWheelListener(transitionHandler);

					newObject.addMouseListener(new AnimationHandler());

				} else if (newObject instanceof Arc) {
					LabelHandler labelHandler = new LabelHandler(((Arc) newObject).getNameLabel(), (Arc) newObject);
					((Arc) newObject).getNameLabel().addMouseListener(labelHandler);
					((Arc) newObject).getNameLabel().addMouseMotionListener(labelHandler);
					((Arc) newObject).getNameLabel().addMouseWheelListener(labelHandler);
					/* CB - Joakim Byg add timed arcs */
					if (newObject instanceof TimedInputArcComponent) {
						if (newObject instanceof TimedTransportArcComponent) {
							TransportArcHandler transportArcHandler = new TransportArcHandler(drawingSurface, (Arc) newObject);
							newObject.addMouseListener(transportArcHandler);
							//newObject.addMouseWheelListener(transportArcHandler);
							newObject.addMouseMotionListener(transportArcHandler);
						} else {
							TimedArcHandler timedArcHandler = new TimedArcHandler(drawingSurface, (Arc) newObject);
							newObject.addMouseListener(timedArcHandler);
							//newObject.addMouseWheelListener(timedArcHandler);
							newObject.addMouseMotionListener(timedArcHandler);
						}
					} else {
						/* EOC */
						ArcHandler arcHandler = new ArcHandler(drawingSurface,(Arc) newObject);
						newObject.addMouseListener(arcHandler);
						//newObject.addMouseWheelListener(arcHandler);
						newObject.addMouseMotionListener(arcHandler);
					}
				} else if (newObject instanceof AnnotationNote) {
					AnnotationNoteHandler noteHandler = new AnnotationNoteHandler(drawingSurface, (AnnotationNote) newObject);
					newObject.addMouseListener(noteHandler);
					newObject.addMouseMotionListener(noteHandler);
					((Note) newObject).getNote().addMouseListener(noteHandler);
					((Note) newObject).getNote().addMouseMotionListener(noteHandler);
				}
				
				newObject.zoomUpdate(drawingSurface.getZoom());
				
			}
			newObject.setGuiModel(template.guiModel());
		}
	}
}
